# -*- coding: utf-8 -*-

from .composition import CompositionWidget
from .data import DataWidget
from .optimization import OptimizationWidget
from .options import OptionsWidget
from .density_optimization import DensityOptimizationWidget
from .extrapolation import ExtrapolationWidget
from .diamond import DiamondWidget
from .configuration import ConfigurationWidget
from .soller import SollerWidget
from .transfer import TransferFunctionWidget